/*
 * $Id: image.c 186 2006-10-05 13:19:42Z csl $
 *
 * Copyright (C) 2006 Christian Stigen Larsen <csl@sublevel3.org>
 * http://csl.sublevel3.org
 *
 * Distributed under the GNU General Public License v2.
 *
 */

#include "image.h"
#include "parse.h"

#ifdef ENABLE_CAIRO
int cairo_get_bpp(const cairo_format_t f) {
	switch ( f ) {
	case CAIRO_FORMAT_ARGB32:    return 4;
	case CAIRO_FORMAT_RGB24:     return 4;
	case CAIRO_FORMAT_A8:        return 1;
	case CAIRO_FORMAT_A1:        // unsupported
	case CAIRO_FORMAT_RGB16_565: // deprecated
	default:
		fputs("error: unsupported output image format\n", stderr);
		exit(1);
	}
}
#endif

const char* image_version_string() {
	static char s[256];
	s[0] = 0;

	#ifdef ENABLE_CAIRO
	sprintf(s, "cairo version: %s", cairo_version_string());
	#endif

	#ifdef ENABLE_GD
	sprintf(s, "gd version: %d\n", GD2_VERS);
	#endif

	return s;
}

void image_new(engine_t *e, const int width_px, const int height_px) {
	#ifdef ENABLE_CAIRO
	e->surface = cairo_image_surface_create(e->output_cairo_format, width_px, height_px);
	e->cr = cairo_create(e->surface);
	cairo_scale(e->cr, width_px, height_px);
	e->cairo_error = 0;
	#endif

	#ifdef ENABLE_GD
	e->im = gdImageCreateTrueColor(width_px, height_px);
	#endif
}

void image_destroy(engine_t *e) {

	#ifdef ENABLE_CAIRO	
	cairo_destroy(e->cr);
	cairo_surface_destroy(e->surface);
	#endif

	#ifdef ENABLE_GD
	gdImageDestroy(e->im);
	#endif
}

const char* image_status_to_string(engine_t *p) {
	if ( p->status_ok )
		return "no error";

	#ifdef ENABLE_CAIRO
	return cairo_status_to_string(p->cairo_error);
	#endif

	#ifdef ENABLE_GD
	return "gd not completely implemented yet";
	#endif
	
	return "current graphics driver not implemented";
}

const char* image_output_format_string(const engine_t* p) {
	#ifdef ENABLE_CAIRO
	switch ( p->output_cairo_format ) {
	case CAIRO_FORMAT_ARGB32:    return "ARGB32";
	case CAIRO_FORMAT_RGB24:     return "RGB24";
	case CAIRO_FORMAT_A8:        return "A8";
	default:                     return "UNKNOWN";
	}
	#endif
	
	#ifdef ENABLE_GD
	switch ( p->format ) {
	case A8:	return "A8";
	case RGB24:	return "RGB24";
	case ARGB32:	return "ARGB32";
	default:	return "UNKNOWN";
	}
	#endif

	return "UNIMPLEMENTED";
}

// return 1 if OK, 0 if error
int image_write(engine_t *e, const char* filename) {

	#ifdef ENABLE_CAIRO
	e->cairo_error = cairo_surface_write_to_png(e->surface, filename);
	e->status_ok = !e->cairo_error;
	#endif

	#ifdef ENABLE_GD
	FILE *f;
	if ( !(f = fopen(filename, "wb")) ) {
		perror("could not open file for writing");
		exit(1);
	}

	gdImagePng(e->im, f);

	fclose(f);
	e->status_ok = 1;
	#endif

	return e->status_ok;
}

void image_resize_crop(engine_t *e, const int width_px, const int height_px) {
	#ifdef ENABLE_CAIRO
	unsigned char *data = cairo_image_surface_get_data(e->surface);
	cairo_format_t format = cairo_image_surface_get_format(e->surface);
	int stride = cairo_get_bpp(e->output_cairo_format) * cairo_image_surface_get_width(e->surface);

	cairo_surface_t *crop = cairo_image_surface_create_for_data(data, format, width_px, height_px, stride);

	// switch current surface to cropped one
	cairo_surface_destroy(e->surface);
	e->surface = crop;
	#endif

	#ifdef ENABLE_GD
	gdImagePtr crop = gdImageCreateTrueColor(width_px, height_px);
	gdImageCopy(crop, e->im, 0, 0, 0, 0, width_px, height_px);
	gdImageDestroy(e->im);
	e->im = crop;
	#endif
}

inline
void image_set_foreground(engine_t *p, const rgb_t* color) {
	#ifdef ENABLE_CAIRO
	static const double alpha = 1.0;
	cairo_set_source_rgba(p->cr, (double) color->r / 255.0f, (double) color->g / 255.0f, (double) color->b / 255.0f, alpha);
	#endif

	#ifdef ENABLE_GD
	static const int alpha = 0;

	if ( (p->col_fg = gdImageColorAllocateAlpha(p->im, color->r, color->g, color->b, alpha)) == -1 ) {
		fputs("error: could not allocate RGBA value for foreground\n", stderr);
		exit(1);
	}
	#endif
}

void image_set_background(engine_t *p, const rgb_t* color) {
	#ifdef ENABLE_CAIRO
	const double alpha = 1.0; // opaque
	cairo_set_source_rgba(p->cr, color->r/255.0f, color->g/255.0f, color->b/255.0f, alpha);
	cairo_rectangle(p->cr, 0.0, 0.0, 1.0, 1.0);
	cairo_fill(p->cr);
	#endif

	#ifdef ENABLE_GD
	const int alpha = 0; // opaque
	p->col_bg = gdImageColorAllocateAlpha(p->im, (int)color->r, (int)color->g, (int)color->b, alpha);

	if ( p->col_bg == -1 ) {
		fputs("error: could not allocate RGBA value for background\n", stderr);
		exit(1);
	}

	gdImageFilledRectangle(p->im, 0, 0, p->im->sx, p->im->sy, p->col_bg);
	#endif
}

void image_set_transparent(engine_t *p) {
	#ifdef ENABLE_CAIRO
	// do nothing
	#endif

	#ifdef ENABLE_GD
	// todo: this does not work as expected, the image doesn't get transparent
        const int alpha = 127; // fully transparent, 7-bit value in GD
        int black = gdImageColorAllocateAlpha(p->im, 0, 0, 0, alpha);

        if ( p->col_bg == -1 ) {
                fputs("error: could not allocate RGBA value for background\n", stderr);
                exit(1);
        }

        gdImageFilledRectangle(p->im, 0, 0, p->im->sx, p->im->sy, black);
	#endif
}

void image_select_font(engine_t *p, const char* fontname, const int bold, const double fontsize) {
	#ifdef ENABLE_CAIRO
	if ( !strcmp(fontname, "default") )
		fontname = "Fixed";

	cairo_select_font_face(p->cr, fontname, CAIRO_FONT_SLANT_NORMAL, bold? CAIRO_FONT_WEIGHT_BOLD : CAIRO_FONT_WEIGHT_NORMAL);
	cairo_set_font_size(p->cr, fontsize);
	#endif

	#ifdef ENABLE_GD
	if ( !strcmp(fontname, "default") )
		fontname = "Courier";
		
	p->bold_font = bold;
	p->fontname = fontname;
	p->fontsize = fontsize * p->font_conversion_factor;
	#endif
}

void image_activate_font(engine_t *p) {
	image_select_font(p, p->fontname, p->bold_font, p->fontsize);
}

image_points_t image_get_font_metrics(engine_t *p, const char* text) {
	#ifdef ENABLE_GD
	int brect[8];
	char *error;
	#endif

	image_points_t r = zero_points;

	#ifdef ENABLE_CAIRO
	cairo_text_extents_t e;
	cairo_text_extents(p->cr, text, &e);
	r.x = e.width;
	r.y = e.height;
	#endif

	#ifdef ENABLE_GD
	brect[0] = brect[1] = brect[2] = brect[3] = brect[4] = brect[5] = brect[6] = brect[7] = 0;

	if ( (error = gdImageStringFT(NULL, &brect[0], 0, (char*) p->fontname, p->fontsize, 0.0, 0, 0, (char*)text)) != NULL ) {
		fprintf(stderr, "error: while getting metrics for font '%s': %s\n", p->fontname, error);

		fprintf(stderr, "\n"
			"If the font was not found, try specifying the exact path to the font you want\n"
			"to use with --font=PATH.  You can also set the font path with the environment\n"
			"variable GDFONTPATH.\n"
			"\n"
			"Current font search path is:\n"
			"\n"
			"%s\n"
			, DEFAULT_FONTPATH);

		exit(1);
	}

	r.x = brect[2] - brect[0];
	r.y = brect[3] - brect[7];
	#endif

	return r;
}

image_points_t image_get_font_max_metrics(engine_t *p, const char* text) {
	char ch[2];
	image_points_t r, m;

	image_reset_points(&r);
	ch[1] = 0;

	while ( *text ) {
		ch[0] = *text++;
		m = image_get_font_metrics(p, ch);
		if ( m.x > r.x ) r.x = m.x;
		if ( m.y > r.y ) r.y = m.y;
	}

	return r;
}

inline
void image_move_to(engine_t *en, const image_point_t x, const image_point_t y) {
	en->pos.x = x;
	en->pos.y = y;

	#ifdef ENABLE_CAIRO
	cairo_move_to(en->cr, en->pos.x, en->pos.y);
	#endif
}

inline
image_points_t image_get_position(engine_t *en) {
	#ifdef ENABLE_CAIRO
	static image_points_t r;
	cairo_get_current_point(en->cr, &r.x, &r.y);
	return r;
	#endif

	#ifdef ENABLE_GD
	return en->pos;
	#endif
}

void image_show_text(engine_t *en, const char* text) {
	#ifdef ENABLE_CAIRO
	cairo_show_text(en->cr, text);
	#endif 

	#ifdef ENABLE_GD
	static int brect[8] = {0, 0, 0, 0, 0, 0, 0, 0};
	static char* error;

	if ( (error = gdImageStringFT(en->im, brect, en->col_fg,
				(char*)en->fontname, en->fontsize, 0.0f,
				en->pos.x, en->pos.y, (char*)text)) != NULL )
	{
		fprintf(stderr, "error: while rendering: %s\n", error);
		exit(1);
	}

	en->pos.x += brect[2] - brect[0];
	#endif
}

void image_putc(engine_t *en, const char c) {
	static image_points_t pos;
	static char text[2] = {0, 0};
	static int tabpos;

	text[0] = c;

	switch ( c ) {
	case '\t':
		for ( tabpos=0; tabpos < en->tab_length; ++tabpos )
			image_putc(en, ' ');
		break;

	case '\n':
		image_move_to(en, zero_point, en->pos.y + en->inc.y);
		break;

	default:
		if ( en->fixed_font )
			pos = image_get_position(en);

		image_show_text(en, text);

		if ( en->fixed_font )
			image_move_to(en, pos.x + en->inc.x, en->pos.y);
		break;
	}
}

void image_fill_rect(engine_t *en, const image_point_t x1, const image_point_t y1, const image_point_t x2, const image_point_t y2) {
	#ifdef ENABLE_CAIRO
	cairo_rectangle(en->cr, x1, y1, x2, y2);
	cairo_fill(en->cr);
	#endif

	#ifdef ENABLE_GD
	gdImageFilledRectangle(en->im, x1, y1, x2, y2, en->col_fg);
	#endif
}

void render_image(engine_t *en, const options_t* const p, FILE *input, image_point_t *xmax, image_point_t *ymax) {

	#ifdef ENABLE_CAIRO
	en->charspacing_px /= p->output_width;
	en->linespacing_px /= p->output_height;
	#endif

	if ( p->output_transparent )
		image_set_transparent(en);
	else
		image_set_background(en, &p->background);

	image_set_foreground(en, &p->foreground);

	image_activate_font(en);

	en->inc = image_get_font_max_metrics(en, "abcdefghijklmnopqrstuvwyxABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789_!*@' ");
	en->inc.x += en->charspacing_px;
	en->inc.y += en->linespacing_px;

	image_reset_points(&en->pos);
	en->pos.y = en->inc.y;

	image_reset_point(xmax);
	image_reset_point(ymax);

	image_move_to(en, en->pos.x, en->pos.y);

	if ( !p->html_input )
		parse_text(en, input, xmax, ymax, 0);
	else
		parse_html(en, input, p, xmax, ymax);
}

void crop_image(options_t *p, engine_t *e, const image_point_t width, const image_point_t height) {
	image_point_t cropx;
	image_point_t cropy;

	#ifdef ENABLE_CAIRO
	cropx = (int) (width  * p->output_width);
	cropy = (int) (height * p->output_height);
	#endif

	#ifdef ENABLE_GD
	cropx = width;
	cropy = height;
	#endif

	if ( cropx > p->output_width )
		cropx = p->output_width;

	if ( cropy > p->output_height ) {

		if ( p->verbose )
			fputs("warning: image output dimensions too small for entire input text\n", stderr);

		cropy = p->output_height;
	}

	if ( p->verbose )
		fprintf(stderr, "cropping image to %dx%d pixels\n", (int) cropx, (int) cropy);

	if ( cropx>0 && cropy>0 )
		image_resize_crop(e, cropx, cropy);
	else {
		if ( p->verbose )
			fputs("warning: will not crop image, because it looks empty\n", stderr);
	}

}

const char* get_output_filename(const options_t *p, const char* filename) {
	static char output[5*1024];
	static const char* def_ext = ".png";

	int file_number_suffix;
	const int MAX_NUMBER_SUFFIX= 999;

	output[0] = 0;

	if ( strlen(filename) + strlen(p->output_file) + strlen(def_ext)
		+ 3  /* numbers 000-999 */ >= sizeof(output)/sizeof(char) )
	{
                fputs("error: not enough  memory for output filename\n", stderr);
                exit(1);
        }

       if ( *p->output_file ) {
                strcpy(output, p->output_file);

                // if e.g. --output=/tmp or --output=/tmp/ then put files under
                // this directory
                if ( isdir(p->output_file) ) {
                        if ( p->output_file[strlen(p->output_file)-1] != '/' )
                                strcat(output, "/");
                        strcat(output, basename(!strcmp(filename, "-") ? "stdin" : filename));
                        remove_ext(output);
                        strcat(output, def_ext);
                }

        } else {
                strcpy(output, basename(!strcmp(filename, "-") ? "stdin" : filename));
                remove_ext(output);
                strcat(output, def_ext);
        }

	// add number suffix if output file already exists
        if ( isfile(output) && !p->output_overwrite ) {
                char *s = (char*) alloca(strlen(output) + 1);
                char *s2 = (char*) alloca(strlen(output) + 1);

                if ( !s || !s2 ) {
                        fputs("error: not enough memory for filename\n", stderr);
                        exit(1);
                }
        
		file_number_suffix=2;
                for (;;) {
                        // use two strings to avoid overlapping memory writes in sprintf
                        strcpy(s2, output);
                        remove_ext(s2);
                        sprintf(s, "%s%d%s", s2, file_number_suffix, def_ext);
                        ++ file_number_suffix;

			if ( file_number_suffix > MAX_NUMBER_SUFFIX ) {
				fprintf(stderr, "error: you already have about %d files already " \
					"with that output name, try using --overwrite\n",
					MAX_NUMBER_SUFFIX);
				exit(1);
			}

			if ( !isfile(s) ) {
				if ( p->verbose )
					fprintf(stderr, "warning: %s exists, consider using --overwrite\n", output);
				strcpy(output, s);
				break;
			}
                }
        }

	return output;
}

void convert_file(options_t *p, engine_t *e, const char* filename) {
	FILE *fin = stdin;
	int status;
	image_point_t xmax, ymax;
	const char* output;

	if ( p->verbose )
		fprintf(stderr, "reading %s\n", !strcmp(filename, "-") ? "standard input" : filename);

	if ( strcmp(filename, "-") ) {
		if ( (fin = fopen(filename, "rb")) == NULL ) {
			perror("error: couldn't open file");
			exit(1);
		}
	}

	image_new(e, p->output_width, p->output_height);

	render_image(e, p, fin, &xmax, &ymax);

	if ( p->output_crop )
		crop_image(p, e, xmax, ymax);

	output = get_output_filename(p, filename);
	fprintf(stderr, "writing %s\n", output);

	status = image_write(e, output);

	image_destroy(e);

	if ( !status ) {
		fprintf(stderr, "error: %s\n", image_status_to_string(e));
		exit(1);
	}
}
