/*
 * $Id: options_t.c 150 2006-09-22 11:39:01Z csl $
 *
 * Copyright (C) 2006 Christian Stigen Larsen <csl@sublevel3.org>
 * http://csl.sublevel3.org
 *
 * Distributed under the GNU General Public License v2.
 *
 */

#include "options_t.h"
#include "image.h" // image_version_string
#include "parse.h" // scan_rgb

const char* name = PACKAGE_NAME;
const char* version = PACKAGE_VERSION;
const char* copyright = "Copyright (C) 2006 Christian Stigen Larsen";
const char* license = "Distributed under the GNU General Public License v2";
const char* bug_email = PACKAGE_BUGREPORT;
const char* project_url = "http://a2png.sourceforge.net";

void set_defaults(options_t *p) {
	p->background.r = p->background.g = p->background.b = 0;
	p->foreground.r = p->foreground.g = p->foreground.b = 255;
	p->html_input = 0;
	p->output_crop = 1;
	p->output_file = "";
	p->output_height = 480;
	p->output_overwrite = 0;
	p->output_transparent = 0;
	p->output_width = 640;
	p->quiet = 0;
	p->verbose = 0;
}

// example:  basename("/tmp/some/where/file.txt") returns "file.txt"
const char* basename(const char* s) {
	const char *end = s + strlen(s);
	while ( end>s && *--end != '/' );
	return *end=='/' ? end+1 : end;
}

// example: remove_ext("/tmp/somefile.png") -> "/tmp/somefile"
void remove_ext(char* s) {
	char *end = s + strlen(s);
	while ( end>s && *--end != '.' );
	if ( *end == '.' ) *end = 0;
}

// example: get_extension("/some/file.jpg") returns "jpg", returns "" if not found
const char* get_extension(const char *filename) {
	const char *end = filename + strlen(filename);
	while ( end>filename && *--end != '.' );
	return *end=='.' ? end + 1 : "";
}

// returns >0 if path points to a directory, 0 for other cases
int isdir(const char* path) {
	struct stat s;
	return !stat(path, &s) ? s.st_mode & S_IFDIR : 0;
}

// returns >0 if path exists and is a regular file
int isfile(const char* path) {
	struct stat s;
	return !stat(path, &s) ? s.st_mode & S_IFREG: 0;
}

void print_version() {
	fprintf(stderr, "%s %s\n", name, version);
	fprintf(stderr, "%s\n%s\n", copyright, license);
	fprintf(stderr, "%s\n", image_version_string());
}

void print_help() {
        fprintf(stderr, "%s %s\n", name, version);
        fprintf(stderr, "%s\n", copyright);
	fprintf(stderr, "%s\n", license);

	fputs(
"\n"
"Usage: a2png [ option(s) ] [ file(s) ]\n"
"\n"
"Converts ASCII text to PNG bitmap images.\n"
"\n"
"OPTIONS\n"
"    -                read text from standard input\n"
"    --background=... set background color in RGB hex values, default is 000000\n"
"    --bold           use bold font\n"
"    --charspacing=.. set spacing between characters in pixels, default is 0\n"
"                     you can use negative values\n"
"    --fixed          use fixed font advancement, regardless of font metrics\n"
"    --font=...       set font family, default is 'Fixed'\n"
"    --font-size=...  set font size (badly implemented), default is 0.025\n"
"    --foreground=... set foreground color in RGB hex values, default is ffffff\n"
"    --format=...     set image pixel format, values can be: argb32, rgb24, a8\n"
"                     default is argb32, a8 is 8 bpp alpha values\n"
"    --height=...     set image height in pixels, default is 480\n"
"    --html-input     read jpa2 --html-raw output\n"
"-h, --help           print help\n"
"    --linespacing=.. set extra line spacing in pixels, default is 2\n"
"                     you can use negative values\n"
"    --no-crop        do not crop output image\n"
"    --output=...     set output filename or directory\n"
"    --overwrite      overwrite output files if they exist, default is to add\n"
"                     number suffix\n"
"-v, --verbose        verbose output\n"
"    --width=...      set image width in pixels, default is 640\n"
"-V, --version        print version\n"
"-s, --silent         do not print anything when running\n"
"    --size=WxH       set image width and height, default is 640x480\n"
"    --tab=...        set number of spaces for tab chars,\n"
"                     the default in the known universe is 8\n"
"\n"
"EXAMPLES\n"
"\n"
"    Convert files to 800x600 pixels:     a2png --size=800x600 file1 file2\n"
"    Convert text from standard input:    a2png --size=800x600 -\n"
"    Convert file with yellow background: a2png --background=ffff00 somefile.txt\n"
"\n"
, stderr),

	fprintf(stderr, "Project homepage on %s\n", project_url);
	fprintf(stderr, "Report bugs to <%s>\n", bug_email);
}

int parse_options(int argc, char** argv, options_t *p, engine_t *en) {
	#define IF_OPT(s1) if ( !strcmp(arg, s1) )
	#define IF_OPTS(s1, s2) if ( !strcmp(arg, s1) || !strcmp(arg, s2) )
	#define IF_VAR(s1) if ( !strncmp(arg, s1, strlen(s1)) )

	char rgb[7];
	int files = 0;
	int n;

	for ( n=1; n < argc; ++n ) {
		const char* arg = argv[n];

		if ( *arg!='-' ) {
			++files;
			continue;
		}

		if ( !strcmp(arg, "-") ) {
			++files;
			continue;
		}

		IF_OPTS("--help", "-h") {
			print_help();
			exit(0);
		}

		IF_OPTS("--version", "-V") {
			print_version();
			exit(0);
		}

		IF_OPT("--fixed") {
			en->fixed_font = 1;
			continue;
		}

		IF_OPTS("--verbose", "-v") {
			p->verbose = 1;
			continue;
		}

		IF_OPT("--html-input") {
			p->html_input = 1;
			continue;
		}

		IF_OPTS("--silent", "-s") {
			p->quiet = 1;
			continue;
		}

		IF_OPT("--transparent") {
			p->output_transparent = 1;
			continue;
		}

		IF_OPT("--crop") {
			fputs("warning: option --crop is deprecated and now default.  see --no-crop\n", stderr);
			continue;
		}

		IF_OPT("--no-crop") {
			p->output_crop = 0;
			continue;
		}

		IF_OPT("--overwrite") {
			p->output_overwrite = 1;
			continue;
		}

		IF_OPT("--bold") {
			en->bold_font = 1;
			continue;
		}

		IF_VAR("--format=") {
			const char *s = arg + 9;

			#ifdef ENABLE_CAIRO
			en->output_cairo_format = -1; // mark as unspecified

			# ifdef HAVE_STRCASECMP
			if ( !strcasecmp(s, "argb32") ) en->output_cairo_format = CAIRO_FORMAT_ARGB32;
			if ( !strcasecmp(s, "rgb24") ) en->output_cairo_format = CAIRO_FORMAT_RGB24;
			if ( !strcasecmp(s, "a8") ) en->output_cairo_format = CAIRO_FORMAT_A8;
			# else
			if ( !strcmp(s, "argb32") || !strcmp(s, "ARGB32") ) en->output_cairo_format = CAIRO_FORMAT_ARGB32;
			if ( !strcmp(s, "rgb24") || !strcmp(s, "RGB24") ) en->output_cairo_format = CAIRO_FORMAT_RGB24;
			if ( !strcmp(s, "a8") || !strcmp(s, "A8") ) en->output_cairo_format = CAIRO_FORMAT_A8;
			# endif
			#endif

			#ifdef ENABLE_GD
			# ifdef HAVE_STRCASECMP
			if ( !strcasecmp(s, "argb32") ) en->format = ARGB32;
			if ( !strcasecmp(s, "rgb24") )  en->format = RGB24;
			if ( !strcasecmp(s, "a8") )     en->format = A8;
			# else
			if ( !strcmp(s, "argb32") || !strcmp(s, "ARGB32") ) en->format = ARGB32;
			if ( !strcmp(s, "rgb24") || !strcmp(s, "RGB24") )   en->format = RGB24;
			if ( !strcmp(s, "a8") || !strcmp(s, "A8") )         en->format = A8;
			# endif
			#endif

			continue;
		}

		IF_VAR("--tab=") {
			en->tab_length = atoi(arg + 6);
			continue;
		}

		IF_VAR("--font=") {
			en->fontname = arg + 7;
			continue;
		}

		IF_VAR("--width=") {
			p->output_width = atoi(arg + 8);
			continue;
		}

		IF_VAR("--output=") {
			p->output_file = arg + 9;
			continue;
		}

		if ( sscanf(arg, "--background=%s", rgb) == 1 ) {
			scan_rgb(rgb, &p->background);
			continue;
		}

		if ( sscanf(arg, "--foreground=%s", rgb) == 1 ) {
			scan_rgb(rgb, &p->foreground);
			continue;
		}

		if ( sscanf(arg, "--size=%dx%d", &p->output_width, &p->output_height) == 2 )
			continue;

		if ( sscanf(arg, "--font-size=%lf", &en->fontsize) == 1 )
			continue;

		IF_VAR("--linespacing=") {
			en->linespacing_px = atoi(arg + 14);
			continue;
		}

		IF_VAR("--charspacing=") {
			en->charspacing_px = atoi(arg + 14);
			continue;
		}

		IF_VAR("--height=") {
			p->output_height = atoi(arg + 9);
			continue;
		}

		fprintf(stderr, "error: unknown option %s\n\n", arg);
		print_help();
		exit(1);
	}

	return files;
}

void check_options(options_t *p, engine_t* en, int files) {

	if ( p->quiet ) { // should be first option to check
		p->verbose = 0;
		fclose(stderr);
	}

	if ( !files ) {
		fputs("error: no input files specified\n\n", stderr),
		print_help();
		exit(1);
	}

	if ( p->output_width < 1 || p->output_height < 1 ) {
		fputs("error: invalid output dimensions\n", stderr);
		exit(1);
	}

	if ( en->fontsize < 0.0001f ) {
		fprintf(stderr, "error: too small font size %f\n", en->fontsize);
		exit(1);
	}

	#ifdef ENABLE_CAIRO
	if ( en->output_cairo_format == -1 ) {
		fputs("error: invalid output format\n\n", stderr);
		exit(1);
	}

	if ( en->output_cairo_format == CAIRO_FORMAT_A8 )
		fputs("warning: CAIRO_FORMAT_A8 is untested, and might generate invalid PNG files\n", stderr);

	if ( p->output_transparent && en->output_cairo_format != CAIRO_FORMAT_ARGB32 ){
		fputs("error: to write transparent images, you must use --format=ARGB32\n", stderr);
		exit(1);
	}
	#endif
}
