/*
 * $Id: parse.c 186 2006-10-05 13:19:42Z csl $
 *
 * Copyright (C) 2006 Christian Stigen Larsen <csl@sublevel3.org>
 * http://csl.sublevel3.org
 *
 * Distributed under the GNU General Public License v2.
 *
 */

#include "parse.h"

// format of s: "rrggbb", as in "4a3f02", color ranges 0--255
inline
void scan_rgb(const char* s, rgb_t* color) {
	#define HEXVAL(c) (c>='0' && c<='9' ? c-'0' : (c>='a' && c<='f' ? c-'a'+10 : (c<='A' && c>='F' ? c-'A'+10 : 0)))

	color->r = (HEXVAL(s[0]) << 4) + HEXVAL(s[1]);
	color->g = (HEXVAL(s[2]) << 4) + HEXVAL(s[3]);
	color->b = (HEXVAL(s[4]) << 4) + HEXVAL(s[5]);
}

// returns 1 if not EOF, 0 if EOF
// see note in parse_html for a rationale of this rather odd tokenizer function
inline
int fgettok(FILE *f, char* buf, size_t buflen) {
	char c;

	// read until we get start of a token
	do {
		if ( (c = fgetc(f)) == EOF ) {
			*buf = 0;
			return 0;
		}

	} while ( c==' ' || c=='\t' || c=='\n' || c=='\r' || c=='<' || c=='>' );

	// read token
	for (;;) {
		if ( c==' ' || c=='\t' || c=='\n' || c=='\r' || c=='<' || c=='>' )
			break;

		if ( --buflen <= 0 )
			break;

		*buf++ = c;
	
		if ( (c = fgetc(f)) == EOF ) {
			*buf = 0;
			return 0;
		}
	}

	*buf = 0;
	return 1;
}

void parse_html(engine_t *en, FILE *f, const options_t * const p, image_point_t *xmax, image_point_t *ymax) {
	/*
	 * WARNING:  This section does NOT contain a HTML parser.  Instead if
	 *           reads only the --html-raw output from jp2a:
	 *
	 *           jp2a --html-raw --colors somepic.jpg | a2png --html-input -
	 *
	 * jp2a is an image-to-ascci converter, see http://jp2a.sourceforge.net
	 *
	 * Input must look *exactly* like this:
	 *
	 *  <span style='color:#aabbcc;'>H</span>
	 *  <span style='color:#aabbcc; background-color:#eeffaa;'>B</span>
         *  <br/>
         *  <span style='color:#ee22aa'>K</span>
         *  ...
	 *
	 * Specifically, the stuff inside style='' must be exactly as shown, with whitespace
	 * and all, to work.  Only the first character in text data is read.
	 *
	 * If you feel like implementing a complete HTML/CSS renderer or implementing KTHML's,
	 * or just want to teach me how to parse, feel free to send me a patch!
	 *
	 */

	char tok[1024];
	rgb_t bg, fg;
	image_points_t pos_before, pos_after;

	#define NEXT_TOKEN if ( !fgettok(f, tok, sizeof(tok)) ) return;

	tok[0] = fgetc(f);
	if ( tok[0] != '<' ) {
		fputs("a2png warning: file does not look like HTML, parsing as text\n", stderr);
		parse_text(en, f, xmax, ymax, tok[0]);
		return;
	}

	for (;;) {

		NEXT_TOKEN;

		if ( !strcmp(tok, "span") ) {

			// reset colors for <span> tag
			bg = p->background;
			fg = p->foreground;

			for (;;) {

				NEXT_TOKEN;

				if ( !strncmp(tok, "style='color:#", 14) ) {
					scan_rgb(tok+14, &fg);
					continue;
				}

				if ( !strncmp(tok, "background-color:#", 18) ) {
					scan_rgb(tok+18, &bg);
					continue;
				}
			
				if ( !strncmp(tok, "/span", 5) )
					break;

				if ( *tok == '&' ) {
					if ( !strcmp(tok, "&apos;") ) *tok = '\'';
					if ( !strcmp(tok, "&quot;") ) *tok = '"';
					if ( !strcmp(tok, "&nbsp;") ) *tok = ' ';
					if ( !strcmp(tok, "&amp;") )  *tok = '&';
					if ( !strcmp(tok, "&lt;") )   *tok = '<';
					if ( !strcmp(tok, "&gt") )    *tok = '>';
				}

				pos_before = image_get_position(en);

				image_set_foreground(en, &fg);
				image_putc(en, *tok);

				pos_after = image_get_position(en);

				if ( pos_after.x > *xmax ) *xmax = pos_after.x;
				if ( pos_after.y > *ymax ) *ymax = pos_after.y;

				if ( !p->output_transparent && bg.r>-1 && bg.g>-1 && bg.b>-1 ) {

					// fill character background
					image_set_foreground(en, &bg);
					image_fill_rect(en, pos_before.x, pos_before.y - en->inc.y, pos_after.x, pos_after.y);

					// redraw original character
					image_move_to(en, pos_before.x, pos_before.y);
					image_set_foreground(en, &fg);
					image_putc(en, *tok);
				}
			}
		}

		if ( !strcmp(tok, "br/") )
			image_putc(en, '\n');
	}
}

void parse_text(engine_t *en, FILE *f, image_point_t *xmax, image_point_t *ymax, char ungetc) {
	char line[10*1024];
	char *p;
	image_points_t cur;

	for (;;) {
		// this is a hack, in parse_html we check the first character,
		// if its not '<' then we continue with this function, and so we
		// must include the first character read.. ungetch() will not work
		// because we can be operating either on stdin or files, therefore
		// this hack
		if ( !ungetc ) {
			if ( !fgets(line, sizeof(line)-1, f) )
				break;
		} else {
			line[0] = ungetc;
			ungetc = 0;
			if ( !fgets(line+1, sizeof(line)-2, f) )
				break;
		}

		// remove newline
		line[strlen(line)-1] = 0;

		// handle carriage returns
		p = line;
		while ( *p ) {
			if ( *p=='\r' ) {
				strcpy(line, p+1);
				p = line;
			} else
				++p;
		}

		image_show_text(en, line);

		cur = image_get_position(en);
		if ( cur.x > *xmax ) *xmax = cur.x;
		if ( cur.y > *ymax ) *ymax = cur.y;
	
		image_putc(en, '\n');
	}
}
